from typing import List,Tuple
import re
import os.path
import pickle
from uuid import uuid4
import sasktran as sk
import numpy as np
import skdoas.retrieval
from skdoas.retrieval.radianceformat import RadianceSKIMPY
from skdoas.retrieval.omno2          import AirMassFactorOMNO2

"""
A set of Doas retrieval objects that are solely built for the Aim North March 2019 test cases. Its not a big job to
make these work for other scenarios but that has not been done yet. 
"""

#------------------------------------------------------------------------------
#           Radiance_AimNorthStudyMar2019
#------------------------------------------------------------------------------

class Radiance_AimNorthStudyMar2019(RadianceSKIMPY):

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self, filepath: str):
        # parse the path to extract profile type, albedo, month, hour, and point-spread information
        filename = os.path.basename(filepath)
        m = re.match('aimnorth_(\w+)_(\d\.\d)_(\d\.\d)_\d+-(\d+)-\d+-(\d+)-\d+-\d+\.pickle', filename)
        if m is None:
            raise ValueError(f'Unrecognized file name pattern {filename}')
        self._ptype = m.group(1)
        self._psf = float(m.group(2))  # nm
        self._alb = float(m.group(3))
        self._month = {'12': 'Dec', '03': 'Mar', '06': 'Jun', '09': 'Sep'}[m.group(4)]
        self._hour = int(m.group(5))

        # save some extra atmosphere information from the file
        with open(filepath, 'rb') as f:
            data = pickle.load(f)
        self._alts = data['profileheights']
        self._no2 = data['no2_profile']
        self._temp = data['temperature']

        super().__init__(skimpy_output_path=filepath)

    @property
    def profile_type(self):
        return self._ptype

    @property
    def psf(self):
        return self._psf

    @property
    def albedo(self):
        return self._alb

    @property
    def month(self):
        return self._month

    @property
    def hour(self):
        return self._hour

    @property
    def profile_heights(self):
        return self._alts

    @property
    def no2_profile(self):
        return self._no2

    @property
    def temperature_profile(self):
        return self._temp

#------------------------------------------------------------------------------
#           Atmosphere_AimNorthStudyMar2019_raw
#------------------------------------------------------------------------------

class Atmosphere_AimNorthStudyMar2019_raw( skdoas.retrieval.Atmosphere):

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self, data_file: str):
        with open(data_file, 'rb') as f:
            self._no2_data = pickle.load(f)

    #------------------------------------------------------------------------------
    #           scalar_property
    #------------------------------------------------------------------------------

    def scalar_property(self, property_name: str, measurement_l1: Radiance_AimNorthStudyMar2019):

        nlos = measurement_l1.observer().shape[0]
        if property_name == 'surface_albedo':
            return measurement_l1.albedo * np.ones(nlos)
        elif property_name == 'tropopause_altitude':
            return 12e3
        else:
            raise ValueError(f'Unknown scalar atmospheric property \'{property_name}\'.')

    #------------------------------------------------------------------------------
    #           profile_property
    #------------------------------------------------------------------------------

    def profile_property(self, property_name: str, measurement_l1: Radiance_AimNorthStudyMar2019, altitude: np.array):
        ground = measurement_l1.ground()  # (nlos, lat-lon-alt)
        nlos = ground.shape[0]

        if property_name == 'no2':
            no2 = np.zeros((nlos, len(altitude)))
            if measurement_l1.profile_type == 'urban':
                no2 += np.interp(altitude, self._no2_data['altitude'][0], self._no2_data['no2_cm3'][0])
            elif measurement_l1.profile_type == 'oilsand':
                no2 += np.interp(altitude, self._no2_data['altitude'][1], self._no2_data['no2_cm3'][1])
            elif measurement_l1.profile_type == 'clean':
                no2 += np.interp(altitude, self._no2_data['altitude'][2], self._no2_data['no2_cm3'][2])
            else:
                raise ValueError(f'Unknown profile type {measurement_l1.profile_type}')
            return no2
        elif property_name == 'temperature':
            temps = np.zeros((nlos, len(altitude)))
            if measurement_l1.profile_type == 'urban':
                temps += np.interp(altitude, self._no2_data['altitude'][0], self._no2_data['temperature'][0])
            elif measurement_l1.profile_type == 'oilsand':
                temps += np.interp(altitude, self._no2_data['altitude'][1], self._no2_data['temperature'][1])
            elif measurement_l1.profile_type == 'clean':
                temps += np.interp(altitude, self._no2_data['altitude'][2], self._no2_data['temperature'][2])
            else:
                raise ValueError(f'Unknown profile type {measurement_l1.profile_type}')
            return temps
        else:
            raise ValueError(f'Unknown atmospheric profile property \'{property_name}\'.')

#------------------------------------------------------------------------------
#           Atmosphere_AimNorthStudyMar2019
#------------------------------------------------------------------------------

class Atmosphere_AimNorthStudyMar2019(skdoas.retrieval.Atmosphere):

    #------------------------------------------------------------------------------
    #           scalar_property
    #------------------------------------------------------------------------------

    def scalar_property(self, property_name: str, measurement_l1: Radiance_AimNorthStudyMar2019):

        nlos = measurement_l1.observer().shape[0]
        if property_name == 'surface_albedo':
            return measurement_l1.albedo * np.ones(nlos)
        elif property_name == 'tropopause_altitude':
            return 12e3
        else:
            raise ValueError(f'Unknown scalar atmospheric property \'{property_name}\'.')

    #------------------------------------------------------------------------------
    #           profile_property
    #------------------------------------------------------------------------------

    def profile_property(self, property_name: str, measurement_l1: Radiance_AimNorthStudyMar2019, altitude: np.array):
        ground = measurement_l1.ground()  # (nlos, lat-lon-alt)
        nlos = ground.shape[0]

        if property_name == 'no2':
            no2 = np.zeros((nlos, len(altitude)))
            no2 += np.interp(altitude, measurement_l1.profile_heights, measurement_l1.no2_profile)
            return no2
        elif property_name == 'temperature':
            temps = np.zeros((nlos, len(altitude)))
            temps += np.interp(altitude, measurement_l1.profile_heights, measurement_l1.temperature_profile)
            return temps
        else:
            raise ValueError(f'Unknown atmospheric profile property \'{property_name}\'.')

#------------------------------------------------------------------------------
#           AMFLookup_AimNorthStudyMar2019
#------------------------------------------------------------------------------

class AMFLookup_AimNorthStudyMar2019(skdoas.retrieval.AirMassFactorProfile):

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self, grid: float, cache_file: str):
        basedir, name = os.path.split(__file__)
        cache_file  = os.path.normpath( basedir + os.sep + cache_file)
        self._cache = cache_file
        self._grid = grid

        if not os.path.isfile(cache_file):                          # initialize the file if it doesn't exist
            profiles = ('clean', 'urban', 'oilsand')
            albedos = (0.0, 0.3, 1.0)
            months = ('Dec', 'Mar', 'Jun', 'Sep')
            hours = tuple(range(0, 24, 2))

            amfs = {}
            for profile in profiles:
                amfs[profile] = {}
                for albedo in albedos:
                    amfs[profile][albedo] = {}
                    for month in months:
                        amfs[profile][albedo][month] = {}
                        for hour in hours:
                            amfs[profile][albedo][month][hour] = None

            with open(cache_file, 'wb') as f:
                pickle.dump(amfs, f)

    #------------------------------------------------------------------------------
    #           air_mass_factor_profile
    #------------------------------------------------------------------------------

    def air_mass_factor_profile(self, measurement_l1: Radiance_AimNorthStudyMar2019,
                                atmosphere: skdoas.retrieval.Atmosphere) -> np.array:
        with open(self._cache, 'rb') as f:
            cache = pickle.load(f)

        p, a, m, h = measurement_l1.profile_type, measurement_l1.albedo, measurement_l1.month, measurement_l1.hour
        if cache[p][a][m][h] is None:
            with open(f'{self._cache}-{str(uuid4())[:8]}.backup', 'wb') as f:
                pickle.dump(cache, f)

            # convert RadianceBase parameters into the correct sk.Geometry object
            obs = measurement_l1.observer()       # (nlos, lat-lon-alt)
            ground = measurement_l1.ground()      # (nlos, lat-lon-alt)
            mjd = np.array([measurement_l1.utc_to_mjd(utc) for utc in measurement_l1.utc()])  # (nlos,)
            nlos = obs.shape[0]

            geometry = sk.Geometry()
            geometry.lines_of_sight = []
            obs_geo = sk.Geodetic()
            ground_geo = sk.Geodetic()
            for i in range(nlos):
                obs_geo.from_lat_lon_alt(*obs[i, :])
                ground_geo.from_lat_lon_alt(*ground[i, :])
                look = ground_geo.location - obs_geo.location
                look *= 1.0 / np.linalg.norm(look)
                geometry.lines_of_sight.append(sk.LineOfSight(mjd[i], obs_geo.location, look))

            atmo = sk.Atmosphere()
            atmo['air'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
            atmo['o3'] = sk.Species(sk.O3DBM(), sk.Labow())
            atmo['no2'] = sk.Species(sk.NO2Vandaele1998(), sk.ClimatologyUserDefined(measurement_l1.profile_heights, {'SKCLIMATOLOGY_NO2_CM3': measurement_l1.no2_profile}))
            atmo.atmospheric_state = sk.MSIS90()
            atmo.brdf = measurement_l1.albedo

            amf_obj = skdoas.AirMassFactorFiniteDifferenceHR(geometry, atmo, 'no2', self.altitude_edges(), progress_bar=True)
            amf_obj.output_format = 'numpy'
            amf = amf_obj([440.])  # shape (nwl (1), nlos, nalt)
            amf = amf[0, :, :]
            cache[p][a][m][h] = amf
            with open(self._cache, 'wb') as f:
                pickle.dump(cache, f)
            return amf
        else:
            amf = cache[p][a][m][h]
            return amf

    #------------------------------------------------------------------------------
    #           altitude_edges
    #------------------------------------------------------------------------------

    def altitude_edges(self) -> np.array:
        return np.arange(0, 8e4, self._grid)


#------------------------------------------------------------------------------
#           AMF_AimNorthStudyMar2019
#------------------------------------------------------------------------------

class AMF_AimNorthStudyMar2019(AirMassFactorOMNO2):

    #------------------------------------------------------------------------------
    #           air_mass_factor
    #------------------------------------------------------------------------------

    def air_mass_factor(self, measurement_l1: skdoas.retrieval.RadianceBase, altitude_bounds: Tuple[float, float]=(None, None)):
        """
        numpy.array : Shape(nlos,)
        """
        amf_profile = self._amf_obj.air_mass_factor_profile(measurement_l1, self._atmosphere)           # (nlos, nalt)
        nlos = amf_profile.shape[0]

        alt_edges = self._amf_obj.altitude_edges()                                                      # (nalt+1,)

        no2_edges = self._atmosphere.profile_property('no2', measurement_l1, alt_edges)          # (nlos, nalt)
        temp_edges = self._atmosphere.profile_property('temperature', measurement_l1, alt_edges)  # (nlos, nalt)
        no2_profile = 50. * (no2_edges[:, :-1] + no2_edges[:, 1:]) * (alt_edges[1:] - alt_edges[:-1])
        temperature = 0.5 * (temp_edges[:, :-1] + temp_edges[:, 1:])

        alpha = 1. - 3e-3 * (temperature - 220.)

        amf = np.sum(no2_profile * amf_profile * alpha, axis=1) / np.sum(no2_profile, axis=1)
        return amf, 0.15 * amf

