from typing import Any,Iterable,Union,Dict
import yaml
import sys
import os
import os.path
import math
import numpy as np
import sasktran as sk
import sasktranif.sasktranif as skif
from skimpy import FOV_Rectangle,Wavelength_GaussianPSF, Pixel,Instrument, ForwardModel, DetectorProperties, Platform, Platform_OnePoint_Nadir
from .detector_hyvisi_chroma import DetectorProperties_Hyvisi_Chroma




#------------------------------------------------------------------------------
#           read_yaml_file
#------------------------------------------------------------------------------

def read_yaml_file(fname: str):
    fullname = __file__                                         # Get the filename of this module
    [basedir, spare] = os.path.split(fullname)                  # split out the list_of_files
    fullname = basedir + os.sep + 'data' + os.sep + fname
    try:
        f = open(fullname, 'rt')
        configuration = yaml.load(f)
        f.close()
    except Exception as inst:
        print(type(inst))  # the exception instance
        print(inst.args)  # arguments stored in .args
        print(inst)  # __str__ allows args to be printed directly,
        raise RuntimeError(
            'Error reading the YAML configuration file %s from file %s' % (fname, fullname,))
    return configuration


#------------------------------------------------------------------------------
#           AimNorth_Simulator
#------------------------------------------------------------------------------

class AimNorth_Simulator(object):
    """
    The following settings are provided below. Defaults are loaded upon initialization. You  just change the values afterwards in your own experiment.

        * aperture_diameter_cms     From aperture diameter in centimeters
        * wavelengths               Thee array or list of wavelengths to calculate
        * psf                       The point spread function of the instrument
        * pixel_iyfov               The horizontal "Y" field of view in radians
        * pixel_izfov               The horizontal "Z" field of view in radians
        * z_icf_bearing             Azimuth o fthe Z direction
        * numhighreslos             Number of high resolution, instantaneous lines of sight in horizontal or vertical direction across field of view (same number used for both Y and Z directions, i.e. total number equals square of this number)
        * numhighreswave            Number of high resolution wavelengths across wavelength bandpass.

    """

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self):
        """
        Creates the Aim-North Simulator object.
        """
        self.platform            : Platform            = Platform_OnePoint_Nadir()
        self.detector            : DetectorProperties  = DetectorProperties_Hyvisi_Chroma()
        self.instrument_settings : Dict[str,Any]       = self._load_default_instrument_settings()


    #------------------------------------------------------------------------------
    #           _set_pixel_center_wavelengths
    #------------------------------------------------------------------------------

    def _set_pixel_center_wavelengths(self, wavelengths: Union[np.ndarray, Iterable]):
        """
        Sets the center wavelengths of the spectrograph pixels. There are two options
        (i) provide an array of wavelengths at whatever points the user chooses or
        (ii) provide an array of 3 elements suitable for numpy.arange (start,stop, stepsize).
        The wavelengths are stored in `instrument_settings[wavelengths]`

        Parameters
        ----------
        wavelengths: np.ndarray
            An array specifying the central wavelength of each pixel used in the spectrograph simulation. It can either be
            just array of wavelengths (nm) or it can be a 3 element array compatible with numpy.arange, in which case we
            check the last element to make sure its less than 10 otehrwise we assume its an array of 3 wavelength points

        """
        w = np.array(wavelengths)                       # get the array of wavelengths passed in
        if (w.size ==3) and (w[2] < 10.0):              # if we have just 3 elements then see if it is settings for arange
            w = np.arange( w[0],w[1],w[2])
        self.instrument_settings['wavelengths'] = w

    #------------------------------------------------------------------------------
    #           _load_default_instrument_settings
    #------------------------------------------------------------------------------

    def _load_default_instrument_settings(self)->Dict[str,Any]:

        self.instrument_settings = {}
        config = read_yaml_file('aim_north_settings.yaml')                                               # read aim-north configuration
        for k,value in config.items():
            self.instrument_settings[k] = value
        w = self.instrument_settings['wavelengths']
        self._set_pixel_center_wavelengths(w)
        return self.instrument_settings

     #------------------------------------------------------------------------------
    #           make_forwardmodel
    #------------------------------------------------------------------------------

    def _make_forwardmodel( self, instrument : Instrument, atmosphere : sk.Atmosphere  )->ForwardModel:
        """
        Makes and configures the sasktran forward model for the Aim North simulations. This typically means configuring
        the atmosphere used for the Sasktran Radiative transfer calculation and setting up the call back function so it
        calls our method to make Level 1 products once radiancesetc. have been calculates.
        """

        model = ForwardModel( level1_supplier_func = self.encode_aim_north_level1_data)
        model.radiance_simulator().set_atmosphere( atmosphere )
        model.radiance_simulator().set_option_weighting_functions_for_toa_albedo( True )
        return model

    #------------------------------------------------------------------------------
    #           make_AimNorth_instrument
    #------------------------------------------------------------------------------

    def _make_instrument( self ):                                                                                        # The number of high resolution wavelengths across the point spread function of each wavelength pixel

        """
        Makes the instrument for Aim-North. This is primarily a matter of defining the field of views of the pixels
        and the wavelength bandpass for various pixels.
        """

        instrument            = Instrument()                                                                            # Create the skimpy instrument object
        aperture_diameter_cms = self.instrument_settings['aperture_diameter_cms']
        wavelengths           = self.instrument_settings['wavelengths']
        psf                   = self.instrument_settings['psf']
        pixel_iyfov           = self.instrument_settings['pixel_iyfov']
        pixel_izfov           = self.instrument_settings['pixel_izfov']
        numhighreslos         = self.instrument_settings['numhighreslos']
        numhighreswave        = self.instrument_settings['numhighreswave']


        fer = instrument.front_end_radiance()

        fov = FOV_Rectangle( Hrange=(numhighreslos, 0.0, pixel_iyfov),
                             Vrange=(numhighreslos, 0.0, pixel_izfov))                                                   # Define AimNorth AQ field of view of 1 pixel. Use just one line of sight for horizontal and vertical
        instrument.front_end_radiance().add_field_of_view( fov)

        wav = list(wavelengths)                                                                                         # Convert the incoming wavelengths into a list
        for w in wav:                                                                                                   # and
            fer.add_wavelength_range( Wavelength_GaussianPSF(numhighreswave, w, psf))                                   # add the wavelength range to our frontendradiance object

        numpix = len(wav)                                                                                               # define all the "spectrum" pixels, one pixel for each wavelength
        for i in range(numpix):                                                                                         # so for each wavelength
            fer.add_pixel( Pixel( fer.fov[0], fer.wavelength_ranges[i]))                                                # create the pixels using the same field of view and different wavelength ranges

        fer.set_aperture_area_cm2( aperture_diameter_cms*aperture_diameter_cms*math.pi/4.0 )                            # Set up the entrace aperture area in cm2

        return instrument

    #------------------------------------------------------------------------------
    #           set_platform_location
    #------------------------------------------------------------------------------

    def set_platform_location(self, latitude:float, longitude:float, utc: np.datetime64,  height_meters:float=37770000.0 ):
        """
        Sets the time and location of the platform using latitude, longitude, time and optional height. Ensures the
        platform is running in nadir mode.

        Parameters
        ----------
        latitude : float
            The latitude of the platform in degrees

        longitude : float
            The longitude of the platform in degrees

        utc : numpy.datetime64
            The UTC time of the measurements.

        height_meters : float
            Optional, default is 37770000.0. The height of the satellite in meters above the surface.

        """
        self.platform.set_platform_location( latlonheightandt= (latitude,longitude,height_meters, utc) )
        self.platform.set_nadir_mode(True)

    #------------------------------------------------------------------------------
    #           set_ground_targets
    #------------------------------------------------------------------------------

    def set_ground_target_locations(self, target_latitudes: np.ndarray, target_longitudes: np.ndarray, clear_states : bool = True ):

        """
        :param target_latitudes:
        :param target_longitudes:
        :param clear_states:
        :return:
        """
        zazi = self.instrument_settings['z_icf_bearing']
        if (clear_states): self.platform.clear_states()
        for i in range(target_latitudes.size):
            self.platform.set_boresight_to_look_at_geodetic_location( target_latitudes[i],
                                                                      target_longitudes[i],
                                                                      0.0,
                                                                      nadir_azimuth_of_inst_z = zazi)
            self.platform.add_current_platform_state()

    #------------------------------------------------------------------------------
    #           calculate_radiance
    #------------------------------------------------------------------------------

    def calculate_radiance( self,
                            atmosphere : sk.Atmosphere,
                            sasktran_options  : Dict[str,Any] = None,
                            sasktranif_options: Dict[str,Any] = None)->Dict[str,Any]:

        instrument   = self._make_instrument()
        forwardmodel = self._make_forwardmodel( instrument, atmosphere )
        #forwardmodel.radiance_simulator().set_option_generate_dummy_data(True)
        if (sasktran_options  ) is not None: forwardmodel.radiance_simulator().set_sasktran_options  ( sasktran_options )
        if (sasktranif_options) is not None: forwardmodel.radiance_simulator().set_sasktranif_options( sasktranif_options)
        level1 = forwardmodel.calculate_radiance( instrument, self.platform)
        return level1


    #------------------------------------------------------------------------------
    #           get_height_profile
    #------------------------------------------------------------------------------

    def get_height_profile( self, species_handle, climatology: sk.Climatology, geodetic_location, heights: np.ndarray  = None):

        climobject = climatology.skif_object()
        location   = list(geodetic_location)
        if heights is None: heights = np.arange(0.0, 80000.0, 500.0)
        values     = np.zeros_like( heights )

        for i in range(heights.size):
            location[2]   = heights[i]
            ok, values[i] = climobject.GetParameter( species_handle, location)
        return heights,values

    #------------------------------------------------------------------------------
    #               encode_aim_north_level1_data(
    #------------------------------------------------------------------------------

    def encode_aim_north_level1_data(   self,
                                        instrument  : Instrument,                       # Radiance information for each observation set sample is stored in the indiviual pixels
                                        forwardmodel: ForwardModel,                     # Forward model used to calculate the radiance,
                                        center_pos  : np.ndarray,                       # GEO Position of platform for each sample Array (3, Nt)
                                        center_los  : np.ndarray,                       # GEO central look vector of each pixel in each sample, Array(3, Numpix, Nt)
                                        utc         : np.ndarray,                       # UTC time of each observation, Array(Nt) of np.datetime64 or possibly float if MJD
                                        center_wave : np.ndarray )->Dict[str,Any]:    # Central wavelength of each pixel in each sample, Array(Numpix,Nt)
        """
        Function encodes the radiance signals into Aim North Level 1 data. This is typically invoked as a callback
        function near the end of instrument.calculate_radiance(). Its purpose is to return a single python object that
        has the Level 1 data for further simulation/retrieval purposes.
        """

        skengine   = forwardmodel.radiance_simulator().engine()
        skgeodetic = sk.Geodetic()
        ok,sun     =  skengine._iskengine.GetProperty('sun')
        ok,refpoint=  skengine._iskengine.GetProperty('referencepoint')

        fer      = instrument.front_end_radiance()
        numwave  = instrument.front_end_radiance().num_pixels()                          # get the number of wavelengths/pixels
        nt       = center_los.shape[2]                                                   # get the number of time samples
        level1_radiance   = np.zeros( (numwave, nt) )
        level1_irradiance = np.zeros( (numwave, nt) )
        level1_toaalbedo  = np.zeros( (numwave, nt) )
        level1_photonflux = np.zeros( (numwave, nt) )
        for iw in range( fer.num_pixels()):
            level1_radiance  [iw, :] = fer.pixels[iw].radiance
            level1_irradiance[iw, :] = fer.pixels[iw].irradiance
            level1_toaalbedo [iw, :] = fer.pixels[iw].toa_albedo
            level1_photonflux[iw, :] = fer.pixels[iw].photon_flux

        satellite_geocentric_pos = np.reshape( center_pos[:,:], (3, nt) )
        satellite_geocentric_los = np.reshape( center_los[:,0,:], (3, nt) )
        satellite_geodetic_pos   = np.zeros( (3,  nt) )
        ground_geodetic_pos      = np.zeros( (3,  nt) )
        ground_sza               = np.zeros( (nt,) )
        ground_ssa               = np.zeros( (nt,) )
        ground_saa               = np.zeros( (nt,) )
        look_zenith              = np.zeros( (nt,) )
        look_azimuth             = np.zeros( (nt,) )
        wavelens                 = np.reshape   ( center_wave[:,0],  (numwave,) )

        atmos = skengine.atmosphere
        state = atmos.atmospheric_state
        no2   = atmos.species['no2'].climatology
        heights, no2profile  = self.get_height_profile( skif.SKCLIMATOLOGY_NO2_CM3,       no2,   refpoint)
        heights, temperature = self.get_height_profile( skif.SKCLIMATOLOGY_TEMPERATURE_K, state, refpoint,        heights=heights)
        heights, airnumber   = self.get_height_profile( skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, state, refpoint, heights = heights)

        for it in range(nt):
            skgeodetic.from_xyz( satellite_geocentric_pos[:,it] )             # Set the satellite position for this sample
            satellite_geodetic_pos[0,it] = skgeodetic.latitude                #
            satellite_geodetic_pos[1,it] = skgeodetic.longitude
            satellite_geodetic_pos[2,it] = skgeodetic.altitude

            intercepts = skgeodetic.altitude_intercepts( 0.0,  satellite_geocentric_pos[:,it], satellite_geocentric_los[:,it])
            ground_point = intercepts[0]
            skgeodetic.from_xyz( ground_point )

            ground_geodetic_pos[0, it] = skgeodetic.latitude
            ground_geodetic_pos[1, it] = skgeodetic.longitude
            ground_geodetic_pos[2, it] = skgeodetic.altitude
            up     = skgeodetic.local_up
            south  = skgeodetic.local_south
            west   = skgeodetic.local_west
            z      = np.dot( up, sun)
            x      = np.dot( sun, -south)
            y      = np.dot( sun, -west )
            xl     = np.dot( -satellite_geocentric_los[:,it], -south)           # Dot of look vector towards observer with North
            yl     = np.dot( -satellite_geocentric_los[:,it], -west )           # Dot of look vector towards observer with East
            coslookzenith = np.dot( -up, satellite_geocentric_los[:,it] )
            cosssa = np.dot( satellite_geocentric_los[:,it], sun )
            ground_saa[it] = math.degrees( math.atan2(y,x) )
            ground_sza[it] = math.degrees( math.acos(z) )
            ground_ssa[it] = math.degrees( math.acos(cosssa) )
            look_zenith[it] = math.degrees( math.acos(coslookzenith) )
            look_azimuth[it] = math.degrees( math.atan2(yl,xl) )

        level1 ={ 'version'                    : '0.2',
                  'radiance'                   : level1_radiance,
                  'irradiance'                 : level1_irradiance,
                  'toa_albedo'                 : level1_toaalbedo,
                  'photon_flux'                : level1_photonflux,
                  'wavelengths'                : wavelens,
                  'utc'                        : utc[0],
                  'satellite_geocentric_pos'   : satellite_geocentric_pos,
                  'satellite_geocentric_los'   : satellite_geocentric_los,
                  'satellite_geodetic_pos'     : satellite_geodetic_pos,
                  'ground_points_geodetic_pos' : ground_geodetic_pos,
                  'sun_vector'                 : sun,
                  'ground_sza'                 : ground_sza,
                  'ground_ssa'                 : ground_ssa,
                  'ground_saa'                 : ground_saa,
                  'look_zenith'                : look_zenith,
                  'look_azimuth'               : look_azimuth,
                  'no2_profile'                : no2profile,
                  'temperature'                : temperature,
                  'airnumberdensity'           : airnumber,
                  'profileheights'             : heights
                  }

        return level1


